@@ -23,16 +23,17 @@
  * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#  include <config.h>
 #endif // HAVE_CONFIG_H
 
 #include <stdio.h>
 #include <string.h>
 #include <CUnit/Basic.h>
 // include test cases' include files here
-#include "shrpx_ssl_test.h"
+#include "shrpx_tls_test.h"
 #include "shrpx_downstream_test.h"
 #include "shrpx_config_test.h"
+#include "shrpx_worker_test.h"
 #include "http2_test.h"
 #include "util_test.h"
 #include "nghttp2_gzip_test.h"
@@ -42,17 +43,19 @@
 #include "shrpx_http_test.h"
 #include "base64_test.h"
 #include "shrpx_config.h"
-#include "ssl.h"
+#include "tls.h"
+#include "shrpx_router_test.h"
+#include "shrpx_log.h"
 
 static int init_suite1(void) { return 0; }
 
 static int clean_suite1(void) { return 0; }
 
 int main(int argc, char *argv[]) {
-  CU_pSuite pSuite = NULL;
+  CU_pSuite pSuite = nullptr;
   unsigned int num_tests_failed;
 
-  nghttp2::ssl::libssl_init();
+  nghttp2::tls::libssl_init();
 
   shrpx::create_config();
 
@@ -62,18 +65,18 @@ int main(int argc, char *argv[]) {
 
   // add a suite to the registry
   pSuite = CU_add_suite("shrpx_TestSuite", init_suite1, clean_suite1);
-  if (NULL == pSuite) {
+  if (nullptr == pSuite) {
     CU_cleanup_registry();
     return CU_get_error();
   }
 
   // add the tests to the suite
-  if (!CU_add_test(pSuite, "ssl_create_lookup_tree",
-                   shrpx::test_shrpx_ssl_create_lookup_tree) ||
-      !CU_add_test(pSuite, "ssl_cert_lookup_tree_add_cert_from_file",
-                   shrpx::test_shrpx_ssl_cert_lookup_tree_add_cert_from_file) ||
-      !CU_add_test(pSuite, "ssl_tls_hostname_match",
-                   shrpx::test_shrpx_ssl_tls_hostname_match) ||
+  if (!CU_add_test(pSuite, "tls_create_lookup_tree",
+                   shrpx::test_shrpx_tls_create_lookup_tree) ||
+      !CU_add_test(pSuite, "tls_cert_lookup_tree_add_ssl_ctx",
+                   shrpx::test_shrpx_tls_cert_lookup_tree_add_ssl_ctx) ||
+      !CU_add_test(pSuite, "tls_tls_hostname_match",
+                   shrpx::test_shrpx_tls_tls_hostname_match) ||
       !CU_add_test(pSuite, "http2_add_header", shrpx::test_http2_add_header) ||
       !CU_add_test(pSuite, "http2_get_header", shrpx::test_http2_get_header) ||
       !CU_add_test(pSuite, "http2_copy_headers_to_nva",
@@ -100,8 +103,10 @@ int main(int argc, char *argv[]) {
                    shrpx::test_http2_get_pure_path_component) ||
       !CU_add_test(pSuite, "http2_construct_push_component",
                    shrpx::test_http2_construct_push_component) ||
-      !CU_add_test(pSuite, "downstream_field_store_add_header_lower",
-                   shrpx::test_downstream_field_store_add_header_lower) ||
+      !CU_add_test(pSuite, "http2_contains_trailers",
+                   shrpx::test_http2_contains_trailers) ||
+      !CU_add_test(pSuite, "downstream_field_store_append_last_header",
+                   shrpx::test_downstream_field_store_append_last_header) ||
       !CU_add_test(pSuite, "downstream_field_store_header",
                    shrpx::test_downstream_field_store_header) ||
       !CU_add_test(pSuite, "downstream_crumble_request_cookie",
@@ -110,6 +115,10 @@ int main(int argc, char *argv[]) {
                    shrpx::test_downstream_assemble_request_cookie) ||
       !CU_add_test(pSuite, "downstream_rewrite_location_response_header",
                    shrpx::test_downstream_rewrite_location_response_header) ||
+      !CU_add_test(pSuite, "downstream_supports_non_final_response",
+                   shrpx::test_downstream_supports_non_final_response) ||
+      !CU_add_test(pSuite, "downstream_find_affinity_cookie",
+                   shrpx::test_downstream_find_affinity_cookie) ||
       !CU_add_test(pSuite, "config_parse_header",
                    shrpx::test_shrpx_config_parse_header) ||
       !CU_add_test(pSuite, "config_parse_log_format",
@@ -118,10 +127,23 @@ int main(int argc, char *argv[]) {
                    shrpx::test_shrpx_config_read_tls_ticket_key_file) ||
       !CU_add_test(pSuite, "config_read_tls_ticket_key_file_aes_256",
                    shrpx::test_shrpx_config_read_tls_ticket_key_file_aes_256) ||
-      !CU_add_test(pSuite, "config_match_downstream_addr_group",
-                   shrpx::test_shrpx_config_match_downstream_addr_group) ||
+      !CU_add_test(pSuite, "worker_match_downstream_addr_group",
+                   shrpx::test_shrpx_worker_match_downstream_addr_group) ||
       !CU_add_test(pSuite, "http_create_forwarded",
                    shrpx::test_shrpx_http_create_forwarded) ||
+      !CU_add_test(pSuite, "http_create_via_header_value",
+                   shrpx::test_shrpx_http_create_via_header_value) ||
+      !CU_add_test(pSuite, "http_create_affinity_cookie",
+                   shrpx::test_shrpx_http_create_affinity_cookie) ||
+      !CU_add_test(pSuite, "http_create_atlsvc_header_field_value",
+                   shrpx::test_shrpx_http_create_altsvc_header_value) ||
+      !CU_add_test(pSuite, "http_check_http_scheme",
+                   shrpx::test_shrpx_http_check_http_scheme) ||
+      !CU_add_test(pSuite, "router_match", shrpx::test_shrpx_router_match) ||
+      !CU_add_test(pSuite, "router_match_wildcard",
+                   shrpx::test_shrpx_router_match_wildcard) ||
+      !CU_add_test(pSuite, "router_match_prefix",
+                   shrpx::test_shrpx_router_match_prefix) ||
       !CU_add_test(pSuite, "util_streq", shrpx::test_util_streq) ||
       !CU_add_test(pSuite, "util_strieq", shrpx::test_util_strieq) ||
       !CU_add_test(pSuite, "util_inp_strlower",
@@ -130,8 +152,6 @@ int main(int argc, char *argv[]) {
       !CU_add_test(pSuite, "util_to_token68", shrpx::test_util_to_token68) ||
       !CU_add_test(pSuite, "util_percent_encode_token",
                    shrpx::test_util_percent_encode_token) ||
-      !CU_add_test(pSuite, "util_percent_encode_path",
-                   shrpx::test_util_percent_encode_path) ||
       !CU_add_test(pSuite, "util_percent_decode",
                    shrpx::test_util_percent_decode) ||
       !CU_add_test(pSuite, "util_quote_string",
@@ -141,6 +161,9 @@ int main(int argc, char *argv[]) {
       !CU_add_test(pSuite, "util_select_h2", shrpx::test_util_select_h2) ||
       !CU_add_test(pSuite, "util_ipv6_numeric_addr",
                    shrpx::test_util_ipv6_numeric_addr) ||
+      !CU_add_test(pSuite, "util_utos", shrpx::test_util_utos) ||
+      !CU_add_test(pSuite, "util_make_string_ref_uint",
+                   shrpx::test_util_make_string_ref_uint) ||
       !CU_add_test(pSuite, "util_utos_unit", shrpx::test_util_utos_unit) ||
       !CU_add_test(pSuite, "util_utos_funit", shrpx::test_util_utos_funit) ||
       !CU_add_test(pSuite, "util_parse_uint_with_unit",
@@ -165,6 +188,19 @@ int main(int argc, char *argv[]) {
                    shrpx::test_util_make_http_hostport) ||
       !CU_add_test(pSuite, "util_make_hostport",
                    shrpx::test_util_make_hostport) ||
+      !CU_add_test(pSuite, "util_strifind", shrpx::test_util_strifind) ||
+      !CU_add_test(pSuite, "util_random_alpha_digit",
+                   shrpx::test_util_random_alpha_digit) ||
+      !CU_add_test(pSuite, "util_format_hex", shrpx::test_util_format_hex) ||
+      !CU_add_test(pSuite, "util_is_hex_string",
+                   shrpx::test_util_is_hex_string) ||
+      !CU_add_test(pSuite, "util_decode_hex", shrpx::test_util_decode_hex) ||
+      !CU_add_test(pSuite, "util_extract_host",
+                   shrpx::test_util_extract_host) ||
+      !CU_add_test(pSuite, "util_split_hostport",
+                   shrpx::test_util_split_hostport) ||
+      !CU_add_test(pSuite, "util_split_str", shrpx::test_util_split_str) ||
+      !CU_add_test(pSuite, "util_rstrip", shrpx::test_util_rstrip) ||
       !CU_add_test(pSuite, "gzip_inflate", test_nghttp2_gzip_inflate) ||
       !CU_add_test(pSuite, "buffer_write", nghttp2::test_buffer_write) ||
       !CU_add_test(pSuite, "pool_recycle", nghttp2::test_pool_recycle) ||
