@@ -36,50 +36,35 @@
 #include <typeinfo>
 #include <algorithm>
 #include <ostream>
+#include <utility>
 
 namespace nghttp2 {
 
-#if __cplusplus > 201103L
-using std::make_unique;
-#else  // __cplusplus <= 201103L
-template <typename T, typename... U>
-typename std::enable_if<!std::is_array<T>::value, std::unique_ptr<T>>::type
-make_unique(U &&... u) {
-  return std::unique_ptr<T>(new T(std::forward<U>(u)...));
-}
-
-template <typename T>
-typename std::enable_if<std::is_array<T>::value, std::unique_ptr<T>>::type
-make_unique(size_t size) {
-  return std::unique_ptr<T>(new typename std::remove_extent<T>::type[size]());
-}
-#endif // __cplusplus <= 201103L
-
 // std::forward is constexpr since C++14
 template <typename... T>
 constexpr std::array<
     typename std::decay<typename std::common_type<T...>::type>::type,
     sizeof...(T)>
-make_array(T &&... t) {
+make_array(T &&...t) {
   return std::array<
       typename std::decay<typename std::common_type<T...>::type>::type,
       sizeof...(T)>{{std::forward<T>(t)...}};
 }
 
-template <typename T, size_t N> constexpr size_t array_size(T(&)[N]) {
+template <typename T, size_t N> constexpr size_t array_size(T (&)[N]) {
   return N;
 }
 
-template <typename T, size_t N> constexpr size_t str_size(T(&)[N]) {
+template <typename T, size_t N> constexpr size_t str_size(T (&)[N]) {
   return N - 1;
 }
 
 // inspired by <http://blog.korfuri.fr/post/go-defer-in-cpp/>, but our
 // template can take functions returning other than void.
 template <typename F, typename... T> struct Defer {
-  Defer(F &&f, T &&... t)
+  Defer(F &&f, T &&...t)
       : f(std::bind(std::forward<F>(f), std::forward<T>(t)...)) {}
-  Defer(Defer &&o) : f(std::move(o.f)) {}
+  Defer(Defer &&o) noexcept : f(std::move(o.f)) {}
   ~Defer() { f(); }
 
   using ResultType = typename std::result_of<typename std::decay<F>::type(
@@ -87,7 +72,7 @@ template <typename F, typename... T> struct Defer {
   std::function<ResultType()> f;
 };
 
-template <typename F, typename... T> Defer<F, T...> defer(F &&f, T &&... t) {
+template <typename F, typename... T> Defer<F, T...> defer(F &&f, T &&...t) {
   return Defer<F, T...>(std::forward<F>(f), std::forward<T>(t)...);
 }
 
@@ -99,27 +84,29 @@ template <typename T, typename F> bool test_flags(T t, F flags) {
 // T *dlnext, which point to previous element and next element in the
 // list respectively.
 template <typename T> struct DList {
-  DList() : head(nullptr), tail(nullptr) {}
+  DList() : head(nullptr), tail(nullptr), len(0) {}
 
   DList(const DList &) = delete;
-
   DList &operator=(const DList &) = delete;
 
-  DList(DList &&other) : head(other.head), tail(other.tail) {
-    other.head = other.tail = nullptr;
-  }
+  DList(DList &&other) noexcept
+      : head{std::exchange(other.head, nullptr)},
+        tail{std::exchange(other.tail, nullptr)},
+        len{std::exchange(other.len, 0)} {}
 
-  DList &operator=(DList &&other) {
+  DList &operator=(DList &&other) noexcept {
     if (this == &other) {
       return *this;
     }
-    head = other.head;
-    tail = other.tail;
-    other.head = other.tail = nullptr;
+    head = std::exchange(other.head, nullptr);
+    tail = std::exchange(other.tail, nullptr);
+    len = std::exchange(other.len, 0);
+
     return *this;
   }
 
   void append(T *t) {
+    ++len;
     if (tail) {
       tail->dlnext = t;
       t->dlprev = tail;
@@ -130,6 +117,7 @@ template <typename T> struct DList {
   }
 
   void remove(T *t) {
+    --len;
     auto p = t->dlprev;
     auto n = t->dlnext;
     if (p) {
@@ -149,7 +137,10 @@ template <typename T> struct DList {
 
   bool empty() const { return head == nullptr; }
 
+  size_t size() const { return len; }
+
   T *head, *tail;
+  size_t len;
 };
 
 template <typename T> void dlist_delete_all(DList<T> &dl) {
@@ -191,7 +182,7 @@ constexpr double operator"" _ms(unsigned long long ms) { return ms / 1000.; }
 // Returns a copy of NULL-terminated string [first, last).
 template <typename InputIt>
 std::unique_ptr<char[]> strcopy(InputIt first, InputIt last) {
-  auto res = make_unique<char[]>(last - first + 1);
+  auto res = std::make_unique<char[]>(last - first + 1);
   *std::copy(first, last, res.get()) = '\0';
   return res;
 }
@@ -241,23 +232,22 @@ public:
   using const_reference = const value_type &;
   using const_pointer = const value_type *;
   using const_iterator = const_pointer;
+  using const_reverse_iterator = std::reverse_iterator<const_iterator>;
 
   ImmutableString() : len(0), base("") {}
   ImmutableString(const char *s, size_t slen)
-      : len(slen), base(copystr(s, len)) {}
-  ImmutableString(const char *s) : len(strlen(s)), base(copystr(s, len)) {}
-  ImmutableString(const std::string &s)
-      : len(s.size()), base(copystr(s.c_str(), s.size())) {}
+      : len(slen), base(copystr(s, s + len)) {}
+  explicit ImmutableString(const char *s)
+      : len(strlen(s)), base(copystr(s, s + len)) {}
+  explicit ImmutableString(const std::string &s)
+      : len(s.size()), base(copystr(std::begin(s), std::end(s))) {}
   template <typename InputIt>
   ImmutableString(InputIt first, InputIt last)
-      : len(std::distance(first, last)), base(copystr(first, len)) {}
+      : len(std::distance(first, last)), base(copystr(first, last)) {}
   ImmutableString(const ImmutableString &other)
-      : len(other.len), base(copystr(other.base, other.len)) {}
-  ImmutableString(ImmutableString &&other) noexcept : len(other.len),
-                                                      base(other.base) {
-    other.len = 0;
-    other.base = "";
-  }
+      : len(other.len), base(copystr(std::begin(other), std::end(other))) {}
+  ImmutableString(ImmutableString &&other) noexcept
+      : len{std::exchange(other.len, 0)}, base{std::exchange(other.base, "")} {}
   ~ImmutableString() {
     if (len) {
       delete[] base;
@@ -272,7 +262,7 @@ public:
       delete[] base;
     }
     len = other.len;
-    base = copystr(other.base, other.len);
+    base = copystr(std::begin(other), std::end(other));
     return *this;
   }
   ImmutableString &operator=(ImmutableString &&other) noexcept {
@@ -282,14 +272,12 @@ public:
     if (len) {
       delete[] base;
     }
-    len = other.len;
-    base = other.base;
-    other.len = 0;
-    other.base = "";
+    len = std::exchange(other.len, 0);
+    base = std::exchange(other.base, "");
     return *this;
   }
 
-  template <size_t N> static ImmutableString from_lit(const char(&s)[N]) {
+  template <size_t N> static ImmutableString from_lit(const char (&s)[N]) {
     return ImmutableString(s, N - 1);
   }
 
@@ -299,18 +287,28 @@ public:
   const_iterator end() const { return base + len; };
   const_iterator cend() const { return base + len; };
 
+  const_reverse_iterator rbegin() const {
+    return const_reverse_iterator{base + len};
+  }
+  const_reverse_iterator crbegin() const {
+    return const_reverse_iterator{base + len};
+  }
+
+  const_reverse_iterator rend() const { return const_reverse_iterator{base}; }
+  const_reverse_iterator crend() const { return const_reverse_iterator{base}; }
+
   const char *c_str() const { return base; }
   size_type size() const { return len; }
   bool empty() const { return len == 0; }
   const_reference operator[](size_type pos) const { return *(base + pos); }
 
 private:
-  const char *copystr(const char *s, size_t slen) {
-    if (slen == 0) {
+  template <typename InputIt> const char *copystr(InputIt first, InputIt last) {
+    if (first == last) {
       return "";
     }
-    auto res = new char[slen + 1];
-    *std::copy_n(s, slen, res) = '\0';
+    auto res = new char[std::distance(first, last) + 1];
+    *std::copy(first, last, res) = '\0';
     return res;
   }
 
@@ -386,24 +384,28 @@ public:
   using const_reference = const value_type &;
   using const_pointer = const value_type *;
   using const_iterator = const_pointer;
+  using const_reverse_iterator = std::reverse_iterator<const_iterator>;
 
   constexpr StringRef() : base(""), len(0) {}
   explicit StringRef(const std::string &s) : base(s.c_str()), len(s.size()) {}
   explicit StringRef(const ImmutableString &s)
       : base(s.c_str()), len(s.size()) {}
-  StringRef(const char *s) : base(s), len(strlen(s)) {}
+  explicit StringRef(const char *s) : base(s), len(strlen(s)) {}
+  constexpr StringRef(const char *s, size_t n) : base(s), len(n) {}
   template <typename CharT>
   constexpr StringRef(const CharT *s, size_t n)
       : base(reinterpret_cast<const char *>(s)), len(n) {}
   template <typename InputIt>
   StringRef(InputIt first, InputIt last)
-      : base(&*first), len(std::distance(first, last)) {}
+      : base(reinterpret_cast<const char *>(&*first)),
+        len(std::distance(first, last)) {}
   template <typename InputIt>
   StringRef(InputIt *first, InputIt *last)
-      : base(first), len(std::distance(first, last)) {}
+      : base(reinterpret_cast<const char *>(first)),
+        len(std::distance(first, last)) {}
   template <typename CharT, size_t N>
-  constexpr static StringRef from_lit(const CharT(&s)[N]) {
-    return StringRef(s, N - 1);
+  constexpr static StringRef from_lit(const CharT (&s)[N]) {
+    return StringRef{s, N - 1};
   }
   static StringRef from_maybe_nullptr(const char *s) {
     if (s == nullptr) {
@@ -413,16 +415,28 @@ public:
     return StringRef(s);
   }
 
-  const_iterator begin() const { return base; };
-  const_iterator cbegin() const { return base; };
+  constexpr const_iterator begin() const { return base; };
+  constexpr const_iterator cbegin() const { return base; };
 
-  const_iterator end() const { return base + len; };
-  const_iterator cend() const { return base + len; };
+  constexpr const_iterator end() const { return base + len; };
+  constexpr const_iterator cend() const { return base + len; };
 
-  const char *c_str() const { return base; }
-  size_type size() const { return len; }
-  bool empty() const { return len == 0; }
-  const_reference operator[](size_type pos) const { return *(base + pos); }
+  const_reverse_iterator rbegin() const {
+    return const_reverse_iterator{base + len};
+  }
+  const_reverse_iterator crbegin() const {
+    return const_reverse_iterator{base + len};
+  }
+
+  const_reverse_iterator rend() const { return const_reverse_iterator{base}; }
+  const_reverse_iterator crend() const { return const_reverse_iterator{base}; }
+
+  constexpr const char *c_str() const { return base; }
+  constexpr size_type size() const { return len; }
+  constexpr bool empty() const { return len == 0; }
+  constexpr const_reference operator[](size_type pos) const {
+    return *(base + pos);
+  }
 
   std::string str() const { return std::string(base, len); }
   const uint8_t *byte() const {
@@ -434,6 +448,11 @@ private:
   size_type len;
 };
 
+inline bool operator==(const StringRef &lhs, const StringRef &rhs) {
+  return lhs.size() == rhs.size() &&
+         std::equal(std::begin(lhs), std::end(lhs), std::begin(rhs));
+}
+
 inline bool operator==(const StringRef &lhs, const std::string &rhs) {
   return lhs.size() == rhs.size() &&
          std::equal(std::begin(lhs), std::end(lhs), std::begin(rhs));
@@ -448,10 +467,23 @@ inline bool operator==(const StringRef &lhs, const char *rhs) {
          std::equal(std::begin(lhs), std::end(lhs), rhs);
 }
 
+inline bool operator==(const StringRef &lhs, const ImmutableString &rhs) {
+  return lhs.size() == rhs.size() &&
+         std::equal(std::begin(lhs), std::end(lhs), std::begin(rhs));
+}
+
+inline bool operator==(const ImmutableString &lhs, const StringRef &rhs) {
+  return rhs == lhs;
+}
+
 inline bool operator==(const char *lhs, const StringRef &rhs) {
   return rhs == lhs;
 }
 
+inline bool operator!=(const StringRef &lhs, const StringRef &rhs) {
+  return !(lhs == rhs);
+}
+
 inline bool operator!=(const StringRef &lhs, const std::string &rhs) {
   return !(lhs == rhs);
 }
@@ -468,6 +500,11 @@ inline bool operator!=(const char *lhs, const StringRef &rhs) {
   return !(rhs == lhs);
 }
 
+inline bool operator<(const StringRef &lhs, const StringRef &rhs) {
+  return std::lexicographical_compare(std::begin(lhs), std::end(lhs),
+                                      std::begin(rhs), std::end(rhs));
+}
+
 inline std::ostream &operator<<(std::ostream &o, const StringRef &s) {
   return o.write(s.c_str(), s.size());
 }
@@ -493,4 +530,19 @@ inline int run_app(std::function<int(int, char **)> app, int argc,
 
 } // namespace nghttp2
 
+namespace std {
+template <> struct hash<nghttp2::StringRef> {
+  std::size_t operator()(const nghttp2::StringRef &s) const noexcept {
+    // 32 bit FNV-1a:
+    // https://tools.ietf.org/html/draft-eastlake-fnv-16#section-6.1.1
+    uint32_t h = 2166136261u;
+    for (auto c : s) {
+      h ^= static_cast<uint8_t>(c);
+      h += (h << 1) + (h << 4) + (h << 7) + (h << 8) + (h << 24);
+    }
+    return h;
+  }
+};
+} // namespace std
+
 #endif // TEMPLATE_H
